import logging
import json
import time
import datetime
import socket
import ssl
import select
import queue
import threading
import sys
import os
from random import randint

from wappsto.object_instantiation import instantiate
from wappsto.object_instantiation import status
from wappsto.connection import send_data
from wappsto.connection import seluxit_rpc
from wappsto.connection import handlers

logging.basicConfig(level=logging.DEBUG,
                    format="%(levelname)s in %(filename)s:%(funcName)s "
                    "at line %(lineno)d occured at "
                    + "%(asctime)s\n\n\t%(message)s\n\n"
                    "Full Path: %(pathname)s\nProcess Name: %(processName)s\n",
                    handlers=[logging.StreamHandler(),
                              logging.FileHandler("./server_logging", mode='w')])
logger = logging.getLogger(__name__)

class FakeSocket:
    """Mock socket implementation class.
    
    Replaces the normal functionality of a socket to allow it to transfer
    information between it's send() and rec() methods through a queue now
    stored on the socket.
    """
    def __init__(self, place_holder1=None, place_holder2=None):
        self.buf = ''
        # SOCK ATTRIBUTES
        self.family = 1
        self.type = 0
        self.proto = 0
        self._num = 0
        self.fileno = self.fileno
        # SOCK ATTRIBUTES

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def connect(self, address=None, port=None):
        self.address = address
        self.port = port
        self.file_to_load = os.path.join(os.path.dirname(__file__),
                                         'test_JSON/62606aea-fc31-4706-8074-83a705fb29e5.json')
        self.instance = instantiate.Instantiator(json_file_name=self.file_to_load, load_from_state_file=False, status=status.Status(), guide=False, path_to_calling_file='')
        self.lock_await = threading.Lock()
        self.lock_thread_await = threading.Lock()
        self.sending_queue = queue.Queue()
        self.rpc = seluxit_rpc.SeluxitRpc(False)
        self.handlers = handlers.Handlers(self.instance)
        self.packet_awaiting_confirm = {}
        self.add_trace_to_report_list = {}
        self.list_of_message_ids = []
        self.connected = True
        self._internal_queue = queue.Queue()
        self._lock = threading.Lock()

    def close(self):
        pass

    def setsockopt(self, SOL_SOCKET, SOCK_OP, place_holder3):
        self.SOL_SOCKET = SOCK_OP
        self.SOCK_OP = SOCK_OP
        self.place_holder3 = place_holder3

    def getsockopt(self, SOL_SOCKET, SOCK_OP):
        return socket.SOCK_STREAM

    def gettimeout(self):
        return 5

    def settimeout(self, temp):
        pass

    def detach(self):
        pass

    def fileno(self):
        self._num += 1
        return self._num

    def reset_list_of_message_ids(self):
        self.list_of_message_ids = []

    def decode_incoming_JSON(self, data):
        return json.loads(data)

    def generate_random_id(self):
        range_start = 10**(12-1)
        range_end = (10**12)-1
        return randint(range_start, range_end)

    def add_id_to_confirm_list(self, data):
        self.lock_await.acquire()
        decoded = json.loads(data.decode('utf-8'))
        self.packet_awaiting_confirm[decoded.get('id')] = decoded
        self.lock_await.release()

    def remove_id_from_confirm_list(self, id):
        self.lock_await.acquire()
        self.packet_awaiting_confirm = self.packet_awaiting_confirm.pop(id)
        self.lock_await.release()

    def incoming_report_request(self, data):
        return_id = data.get('id')
        try:
            get_url_id = data.get('params').get('url')
        except AttributeError as e:
            error_str = 'Error received incorrect format in get'
            msg = "Report Request from url ID: {}".format(get_url_id)
            logger.info(msg)
            return self.handle_incoming_error(data, e, error_str, return_id)

        try:
            trace_id = data.get('params').get('meta').get('trace')
            logger.info("Report GET found trace id: {}"
                               .format(trace_id))
        except AttributeError:
            trace_id = None

        network_id = None
        device_id = None
        value_id = None
        state_id = None
        result = get_url_id.split('/')
        for item in result:
            if item == 'network':
                network_id = result[2]
            if item == 'device':
                device_id = result[4]
            if item == 'value':
                value_id = result[6]
            if item == 'state':
                state_id = result[8]

        self.build_get_state_reply(data, state_id)
        
        # else:
        #     error = 'Non-existing ID for get'
        #     self.send_error(error, return_id)

    def build_get_state_reply(self, data, state_id):
        state_type = None
        state_data = None
        for device in self.instance.device_list:
            for value in device.value_list:
                if value.report_state is not None:
                    if value.report_state.uuid == state_id:
                        state_type = value.report_state.state_type
                        state_data = value.init_value
                if value.control_state is not None:
                    if value.control_state.uuid == state_id:
                        state_type = value.control_state.state_type
                        state_data = value.init_value

        message_id = 1
        request_id = data.get('params').get('data').get('meta').get('id')
        manufacturer_id = '31439b87-040b-4b41-b5b8-f3774b2a1c19'
        timestamp = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        created_timestamp = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        updated_timestamp = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        server_send_time = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        reply_base = '{"jsonrpc":"2.0","id":' + str(message_id) + ',"result":{"value":{"meta":{"type":"state","version":"2.0","id":"' + request_id + '","manufacturer":"' + manufacturer_id + '","updated":"' + updated_timestamp + '","created":"'+ created_timestamp +'","revision":1,"contract":[]},"type":"' + state_type + '", "status":"Send","data":"' + state_data + '","timestamp":"' + timestamp + '"},"meta":{"server_send_time":"' + server_send_time + '"}' + '}' + '}'
        self.sending_queue.put(bytes(reply_base, 'utf-8'))

    def send_trace(self, sending_queue, parent, trace_id, data, control_value_id=None):
        pass

    def incoming_control(self, data):
        return_id = data.get('id')
        try:
            control_id = data.get('params').get('data').get('meta').get('id')
        except AttributeError as e:
            error_str = 'Error received incorrect format in put'
            return self.handle_incoming_error(data, e, error_str, return_id)
        logger.info("Control Request from control id: " + control_id)

        try:
            local_data = data.get('params').get('data')
        except AttributeError:
            error = 'Error received incorrect format in put, data missing'
            self.send_error(error, return_id)
            return
        try:
            trace_id = data.get('params').get('meta').get('trace')
            logger.info("Control found trace id: " + trace_id)
        except AttributeError:
            # ignore
            trace_id = None
        if self.handle_incoming_put(
                control_id,
                local_data,
                self.sending_queue,
                trace_id
        ):
            self.send_success_reply(return_id)
        else:
            error = 'Invalid value range or non-existing ID'
            self.send_error(error, return_id)

    def validate_initialization(self, data):

        if data['params']['url'].endswith('/network'):
            if data['params']['data']['meta']['id'] == self.instance.network_cl.uuid:
                self.send_success_reply(self.rpc.get_next_random_id())
        for device in self.instance.device_list:
            if data['params']['url'].endswith('/device'):
                if data['params']['data']['meta']['id'] == device.uuid:
                    self.send_success_reply(self.rpc.get_next_random_id())

            for value in device.value_list:
                if data['params']['url'].endswith('/value/'):
                    if data['params']['data']['meta']['id'] == value.uuid:
                        self.send_success_reply(self.rpc.get_next_random_id())

                elif data['params']['url'].endswith('/state'):
                    if value.report_state is not None:
                        if data['params']['data']['meta']['id'] == value.report_state.uuid:
                            self.send_success_reply(self.rpc.get_next_random_id())
                    if value.control_state is not None:
                        if data['params']['data']['meta']['id'] == value.control_state.uuid:
                            self.send_success_reply(self.rpc.get_next_random_id())
        if (not data['params']['url'].endswith('/network')
            and not data['params']['url'].endswith('/device')
            and not data['params']['url'].endswith('/value/')
            and not data['params']['url'].endswith('/state')):
            self.send_error('Could not initialize', self.rpc.get_next_random_id())

    def create_outgoing_put(self):
        self.OUTGOING_PUT = True

    def send_error(self, error_str, return_id):
        error_reply = send_data.SendData(
            send_data.SEND_FAILED,
            rpc_id=return_id,
            text=error_str
        )
        self.sending_queue.put(error_reply)

    def send_success(self, package):
        try:
            rpc_success_response = self.rpc.get_rpc_success_response(
                package.rpc_id
            )
            return rpc_success_response
            logger.info("Sending Successful")

        except OSError as e:
            self.connected = False
            msg = "Error sending response: {}".format(e)
            logger.error(msg, exc_info=True)

    def send_success_reply(self, return_id):
        success_reply = send_data.SendData(
            send_data.SEND_SUCCESS,
            rpc_id=return_id
        )
        self.sending_queue.put(success_reply)

    def send_failed(self, package):
        logger.info("Sending Error")
        rpc_fail_response = self.rpc.get_rpc_fail_response(
            package.rpc_id,
            package.text
        )
        logger.info(rpc_fail_response)
        try:
            return rpc_fail_response
        except OSError as e:
            self.connected = False
            msg = "Error sending failed response: {}".format(e)
            logger.error(msg, exc_info=True)

    def send_report(self, package):
        try:
            if not package.trace_id:
                if package.value_id in self.add_trace_to_report_list.keys():
                    package.trace_id = (
                        self.add_trace_to_report_list.pop(package.value_id)
                    )
            local_data = self.rpc.get_rpc_state(
                package.data,
                package.network_id,
                package.device_id,
                package.value_id,
                package.state_id,
                'report',
                trace_id=package.trace_id
            )
            self.add_id_to_confirm_list(local_data)
            return local_data
            decoded = json.loads(local_data.decode('utf-8'))
            data_decoded = decoded.get('params').get('data').get('data')
            logger.info('Sending report value: {}'.format(data_decoded))
        except OSError as e:
            self.connected = False
            msg = "Error sending report: {}".format(e)
            logger.error(msg, exc_info=True)

    def send_control(self, package):
        logger.info("Sending control message")
        try:
            local_data = self.rpc.get_rpc_state(
                package.data,
                package.network_id,
                package.device_id,
                package.value_id,
                package.state_id,
                'control',
                trace_id=package.trace_id
            )
            self.add_id_to_confirm_list(local_data)
            return local_data
        except OSError as e:
            self.connected = False
            msg = "Error sending control: {}".format(e)
            logger.error(msg, exc_info=True)

    def refresh_button(self, name_of_value):
        for device in self.instance.device_list:
            a_value = device.get_value(name_of_value)
            if a_value.report_state is not None:
                server_send_time = datetime.datetime.utcnow().strftime(
                    '%Y-%m-%dT%H:%M:%S.%fZ')
                a_value.report_state.timestamp = server_send_time
                logger.critical(a_value.report_state.timestamp)
                reply_base = '{"jsonrpc":"2.0","id":"'+str(self.rpc.get_next_random_id())+'","method":"GET","params":{"url":"/network/'+str(device.get_parent_network().uuid)+'/device/'+str(device.uuid)+'/value/'+str(a_value.uuid)+'/state/'+str(a_value.report_state.uuid)+'","data":""}}'
                self.add_id_to_confirm_list(bytes(reply_base, 'utf-8'))
                self.sending_queue.put(bytes(reply_base, 'utf-8'))
                logger.critical("Is Empty: {}".format(self.sending_queue.empty()))
                w = json.loads(reply_base).get('params').get('url')
                self.handlers.handle_incoming_get(w, self.sending_queue, False)


    def handle_incoming_put(self, control_id, local_data, sending_queue,
                            trace_id):
        for device in self.instance.device_list:
            for value in device.value_list:
                if local_data['type'] == 'Report':
                    if value.report_state is not None:
                        if value.report_state.uuid == local_data['meta']['id']:
                            value.last_update_of_report = local_data['timestamp']
                            value.data_value = local_data['data']
                            return True
                if local_data['type'] == 'Control':
                    if value.control_state is not None:
                        if value.control_state.uuid == local_data['meta']['id']:
                            value.last_update_of_control = local_data[
                                'timestamp']
                            value.data_value = local_data['data']
                            return True

    def send(self, data):
        self._lock.acquire()
        if self.connected:
            if data is not None:
                decoded = json.loads(data.decode('utf-8'))
                decoded_id = decoded.get('id')
                self.list_of_message_ids.append(decoded_id)
                try:
                    logger.debug('Raw received Json: {}'.format(decoded))
                    if decoded.get('method', False) == 'PUT':
                        self.incoming_control(decoded)

                    elif decoded.get('method', False) == 'GET':
                        self.incoming_report_request(decoded)

                    elif decoded.get('method', False) == 'POST':
                        self.validate_initialization(decoded)

                    elif decoded.get('error', False):
                        decoded_error = decoded.get('error')
                        msg = "Error: {}".format(decoded_error.get('message'))
                        logger.error(msg)
                        self.remove_id_from_confirm_list(decoded_id)

                    elif decoded.get('result', False):
                        msg = "Successful reply for id {}".format(decoded_id)
                        logger.debug(msg)
                        self.remove_id_from_confirm_list(decoded_id)

                    else:
                        logger.info("Unhandled method")
                        error_str = 'Unknown method'
                        self.send_error(error_str, decoded_id)

                except ValueError:
                    logger.info("Value error")
                    logger.info(data)
                    error_str = 'Value error'
                    self.send_error(error_str, decoded_id)
        else:
            raise Exception
        self._lock.release()

    def recv(self, limit=4096):
        while True:
            if self.connected:
                if not self.sending_queue.empty():
                    package = self.sending_queue.get_nowait()
                    if type(package) == bytes:
                        return bytes(package, 'utf-8')

                    elif package.msg_id == send_data.SEND_SUCCESS:
                        return self.send_success(package)

                    elif package.msg_id == send_data.SEND_REPORT:
                        return self.send_report(package)

                    elif package.msg_id == send_data.SEND_FAILED:
                        return self.send_failed(package)

                    elif package.msg_id == send_data.SEND_CONTROL:
                        return self.send_control(package)

                    elif package.msg_id == send_data.SEND_TRACE:
                        pass

                    else:
                        logger.info("Unhandled send")

                    self.sending_queue.task_done()
            else:
                raise Exception


#!/usr/bin/env python3
import os
import sys
import time
import json
import pytest
import socket
import wappsto
import datetime
from mock import Mock
from test import fake_socket
from test import mock_wappsto_server
from wappsto.connection.seluxit_rpc import SeluxitRpc
from wappsto.connection.network_classes.state import State
from wappsto.connection.network_classes.device import Device
from wappsto.connection.network_classes.errors.wappsto_errors import ServerConnectionException as ServerConnectionException
from wappsto.connection.network_classes.errors.wappsto_errors import CallbackNotCallableException as CallbackNotCallableException
# Bypassing top-level import issues
from pathlib import Path
file = Path(__file__).resolve()
parent, root = file.parent, file.parents[1]
sys.path.append(str(root))
try:
    sys.path.remove(str(parent))
except ValueError:
    pass




def fake_set(self):
    self.wapp_log.info("Totally SSL wrapped the socket. Totally.")
    return self.my_raw_socket


class TestInstantiation:

    def setup_method(self):
        self.json_ok_two_devices = os.path.join(
            os.path.dirname(__file__),
            'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce.json'
        )
        self.json_incorrect = os.path.join(
            os.path.dirname(__file__),
            'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce_invalid.json'
        )

    def test_instantiation_device_list_passes(self):

        assert len(wappsto.Wappsto(json_file_name=self.json_ok_two_devices).instance.device_list) > 0

    def test_instantiation_device_list_2_devices_passes(self):

        assert len(wappsto.Wappsto(json_file_name=self.json_ok_two_devices).instance.device_list) == 2

    def test_instantiation_parsing_fails(self):

        with pytest.raises(json.decoder.JSONDecodeError):
            wappsto.Wappsto(json_file_name=self.json_incorrect)

    def test_instantiation_parsing_invalid_JSON_errors(self):
        with pytest.raises(Exception):
            wappsto.Wappsto(json_file_name=self.json_incorrect)

    def test_instantiation_parsing_file_not_found_errors(self):

        with pytest.raises(FileNotFoundError):
            wappsto.Wappsto(json_file_name="blabla")


class TestDevice:

    @classmethod
    def setup_method(self):
        self.json_ok_two_devices = os.path.join(os.path.dirname(__file__), 'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce.json')
        self.json_incorrect = os.path.join(os.path.dirname(__file__), 'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce_invalid.json')
        self.wapp = wappsto.Wappsto(json_file_name=self.json_ok_two_devices)

    def test_finding_device_passes(self):
        device = self.wapp.get_device("test_device_all_fieldsasdasd")
        assert device is not None

    def test_finding_device_errors(self):
        self.test_wapp = wappsto.Wappsto(json_file_name=self.json_ok_two_devices)
        with pytest.raises(Exception):
            self.test_wapp.get_device("asdfasf")

    def test_set_value_callback_passes(self):
        device = self.wapp.get_device("test_device_all_fieldsasdasd")
        value = device.get_value("test_value_with_number")
        def value_callback(self):
            pass
        value.set_callback(value_callback)
        assert value.callback == value_callback

    def test_set_value_callback_with_return_passes(self):
        device = self.wapp.get_device("test_device_all_fieldsasdasd")
        value = device.get_value("test_value_with_number")
        def value_callback(self):
            return 3
        value.set_callback(value_callback)
        assert value.callback(self) == 3


class TestValue:

    @classmethod
    def setup_method(self):
        self.json_ok_two_devices = os.path.join(os.path.dirname(__file__), 'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce.json')
        self.wapp = wappsto.Wappsto(json_file_name=self.json_ok_two_devices)

    def test_finding_value_passes(self):
        device = self.wapp.get_device("test_device_all_fieldsasdasd")
        value = device.get_value("test_value_with_number")
        assert value is not None

    def test_device1_value_length_is_3_passes(self):
        device1 = self.wapp.get_device("test_device_all_fieldsasdasd")
        assert len(device1.value_list) == 3

    def test_device2_value_length_is_3_passes(self):
        device2 = self.wapp.get_device("test_device_not_all_fields")
        assert len(device2.value_list) == 3

    def test_value_wrong_type_of_value_error_logs(self, caplog):
        value_to_test = None
        for device in self.wapp.instance.device_list:
            for value in device.value_list:
                if value.control_state is not None:
                    if value.number_min is not None:
                        value_to_test = value
        value_to_test.send_control('asd')

        assert 'Invalid type of value. Must be a number.' in caplog.text

    def test_value_string_over_max_error_logs(self, caplog):
        value_to_test = None
        for device in self.wapp.instance.device_list:
            for value in device.value_list:
                if value.control_state is not None:
                    if value.string_max is not None:
                        if value.string_max < 3:
                            value_to_test = value
        msg = 'asd'
        value_to_test.send_control(msg)

        assert 'Value {} not in correct range for {}'.format(msg, value_to_test.name) in caplog.text

    def test_set_not_callable_callback_errors(self, caplog):
         for device in self.wapp.instance.device_list:
            for value in device.value_list:
                if value.name == 'test_value_with_number':
                    with pytest.raises(CallbackNotCallableException):
                        value.set_callback(3)


class TestState:
    @classmethod
    def setup_method(self):
        self.json_ok_two_devices = os.path.join(os.path.dirname(__file__), 'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce.json')
        self.wapp = wappsto.Wappsto(json_file_name=self.json_ok_two_devices)

    def test_finding_state_passes(self):
        device = self.wapp.get_device("test_device_all_fieldsasdasd")
        value = device.get_value("test_value_blob")

        assert value.get_report_state() is not None and value.get_control_state() is not None


class TestStatus:

    @classmethod
    def setup_method(self):
        self.json_ok_two_devices = os.path.join(os.path.dirname(__file__), 'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce.json')
        self.json_incorrect = os.path.join(os.path.dirname(__file__), 'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce_invalid.json')

        self.wapp = wappsto.Wappsto(json_file_name=self.json_ok_two_devices)

    def test_get_status_reference_passes(self):
        status = self.wapp.get_status()

        assert status is not None

    def test_set_status_callback_passes(self):
        status = self.wapp.get_status()

        def status_callback(self):
            pass
        status.set_callback(status_callback)

        assert status.callback is status_callback


class TestSendReceive:

    @pytest.fixture(autouse=True)
    def mock_server_sockets(self, mocker):
        from wappsto.connection.communication import ClientSocket
        mocker.patch.object(socket, 'socket', new=fake_socket.FakeSocket)
        mocker.patch.object(ClientSocket, 'ssl_wrap', new=fake_set)
        mocker.patch.object(ClientSocket, 'init_ok', new=lambda f: True)

    @classmethod
    def setup_class(self):
        one = 'test_JSON/62606aea-fc31-4706-8074-83a705fb29e5.json'
        two = 'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce.json'
        three = 'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce_invalid.json'
        self.json_value_one = os.path.join(os.path.dirname(__file__),one)
        self.json_ok_two_devices = os.path.join(os.path.dirname(__file__),two)
        self.json_incorrect = os.path.join(os.path.dirname(__file__), three)
        self.wapp = wappsto.Wappsto(json_file_name=self.json_value_one)

    def test_wappsto_send_receive_response_passes(self, mocker):
        self.wapp.start(address="127.0.0.1", port=8080)
        message_ids = self.wapp.socket.my_socket.list_of_message_ids
        assert len(message_ids) == 15

    def test_wappsto_refresh_button_changes_updates_timestamp_passes(self):
        self.wapp.start(address="127.0.0.1", port=8080)
        device = self.wapp.get_device("Lamp")
        time1 = device.get_value("ON/OFF").last_update_of_report

        def w(value, type):
            if type == "report":
                x = "It is a report callback"
            else:
                x = "It is a control callback"

        device.get_value("ON/OFF").set_callback(w)
        time.sleep(2)
        self.wapp.socket.my_socket.refresh_button("ON/OFF")
        device = self.wapp.get_device("Lamp")
        time2 = device.get_value("ON/OFF").report_state.timestamp
        self.clear_logs()
        assert time1 is not time2

    def test_wappsto_delta_test_first(self):
        self.wapp.start(address="127.0.0.1", port=8080)
        delta = 0.1
        list = []
        values_list = [0, 1, 1, 0, 2]
        value = self.wapp.get_device("Lamp").get_value("ON/OFF")
        value.set_delta(delta)
        for i in range(len(values_list)):
            value.last_controlled = values_list[i]
            value_prepared = value.last_update_of_report[11:-8]
            list.append(datetime.datetime.strptime(value_prepared, '%H:%M:%S'))
            time.sleep(2)
        time.sleep(1)
        self.clear_logs()
        control_set_length = set(list)
        assert len(control_set_length) == 4

    def test_wappsto_delta_test_second(self):
        self.wapp.start(address="127.0.0.1", port=8080)
        delta = 100
        list = []
        values_list = [0, 1, 1, 1, 0]
        value = self.wapp.get_device("Lamp").get_value("ON/OFF")
        value.set_delta(delta)
        for i in range(len(values_list)):
            value.last_controlled = values_list[i]
            value_prepared = value.last_update_of_report[11:-8]
            list.append(datetime.datetime.strptime(value_prepared, '%H:%M:%S'))
            time.sleep(2)
        time.sleep(1)
        self.clear_logs()
        control_set_length = set(list)
        assert len(control_set_length) == 1

    def test_wappsto_period_test(self):
        self.wapp.start(address="127.0.0.1", port=8080)
        refreshment_time = 2
        assert_count = 0
        value = self.wapp.get_device("Lamp").get_value("ON/OFF")
        value.set_period(refreshment_time)
        for i in range(refreshment_time):
            time.sleep(3)
        file = open("server_logging", "r")
        for line in file:
            if "Sending report [PERIOD]." in line:
                assert_count = assert_count + 1
        self.clear_logs()
        assert assert_count in [3, 4]

    def test_wappsto_period_test_with_refreshment(self):
        assert_count = 0
        self.wapp.start(address="127.0.0.1", port=8080)
        refreshment_time = 3
        value = self.wapp.get_device("Lamp").get_value("ON/OFF")
        value.set_period(refreshment_time)
        for i in range(2):
            self.wapp.socket.my_socket.refresh_button("ON/OFF")
            time.sleep(4)
        time.sleep(1)
        file = open("server_logging", "r")
        for line in file:
            if "Sending report [PERIOD]." in line:
                assert_count = assert_count + 1
        self.clear_logs()

        assert assert_count == 3

    def test_wappsto_period_test_with_refreshment_no_period_refresh(self):
        assert_count = 0
        self.wapp.start(address="127.0.0.1", port=8080)
        refreshment_time = 10
        value = self.wapp.get_device("Lamp").get_value("ON/OFF")
        value.set_period(refreshment_time)
        for i in range(3):
            self.wapp.socket.my_socket.refresh_button("ON/OFF")
            time.sleep(2)
        time.sleep(1)
        file = open("server_logging", "r")
        for line in file:
            if "Sending report [PERIOD]." in line:
                assert_count = assert_count + 1
        self.clear_logs()
        assert assert_count == 0

    def test_wappsto_value_incorrect_range(self):
        self.wapp.start(address="127.0.0.1", port=8080)
        value = self.wapp.get_device("Lamp").get_value("ON/OFF")
        value.send_control(8)
        time.sleep(1)
        assert_count = 0
        file = open("server_logging", "r")
        for line in file:
            if "Invalid number. Range:" in line:
                assert_count = assert_count + 1
        self.clear_logs()
        assert assert_count == 1

    def test_wappsto_value_correct_range(self):
        self.wapp.start(address="127.0.0.1", port=8080)
        value = self.wapp.get_device("Lamp").get_value("ON/OFF")
        value.send_control(1)
        time.sleep(0.5)
        assert value.data_value == 1

    def test_wappsto_set_delta_for_string_value(self):
        self.wapp.start(address="127.0.0.1", port=8080)
        delta = 10
        value = self.wapp.get_device("Lamp").get_value("Displaying text")
        value.set_delta(delta)
        result = False
        file = open("server_logging", "r")
        for line in file:
            if "Cannot set the delta for this value." in line:
                result = True
        assert result is True
        self.clear_logs()


    def test_wappsto_set_incorrect_delta_type(self):
        self.wapp.start(address="127.0.0.1", port=8080)
        delta = "It is not a number"
        value = self.wapp.get_device("Lamp").get_value("ON/OFF")
        value.set_delta(delta)
        result = False
        file = open("server_logging", "r")
        for line in file:
            if "Delta value must be a number" in line:
                result = True
        assert result is True
        self.clear_logs()

    def test_wappsto_set_period_for_value_without_report_state(self):
        self.wapp.start(address="127.0.0.1", port=8080)
        period = 4
        value = self.wapp.get_device("Lamp").get_value("ON/OFF")
        value.report_state = None
        value.set_period(period)
        result = False
        file = open("server_logging", "r")
        for line in file:
            if "Cannot set the period for this value." in line:
                result = True
        assert result is True
        self.clear_logs()

    def clear_logs(self):
        open("server_logging", "w").close()


class TestSaveObjectInstances:

    @pytest.fixture(autouse=True)
    def clean_files(self):
        folder = os.path.join(os.path.dirname(__file__), 'saved_instances/')
        if os.path.exists(folder):
            for f in os.listdir(folder):
                if os.path.isfile(os.path.join(folder, f)):
                    os.remove(os.path.join(folder, f))

    @pytest.fixture(autouse=True)
    def mock_server_sockets(self, mocker):
        from wappsto.connection.communication import ClientSocket
        mocker.patch.object(socket, 'socket', new=fake_socket.FakeSocket)
        mocker.patch.object(ClientSocket, 'ssl_wrap', new=fake_set)
        mocker.patch.object(ClientSocket, 'init_ok', new=lambda f: True)

    @classmethod
    def setup_class(self):
        one = 'test_JSON/62606aea-fc31-4706-8074-83a705fb29e5.json'
        two = 'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce.json'
        three = 'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce_invalid.json'
        self.json_value_one = os.path.join(os.path.dirname(__file__), one)
        self.json_ok_two_devices = os.path.join(os.path.dirname(__file__), two)
        self.json_incorrect = os.path.join(os.path.dirname(__file__), three)
        self.wapp = wappsto.Wappsto(json_file_name=self.json_value_one)
        self.path_to_saved = os.path.join(self.wapp.path_to_calling_file, 'saved_instances/')

    def test_if_file_not_empty_passes(self, capsys):
        self.wapp.start(address="127.0.0.1", port=8080)
        self.wapp.stop()

        result = None
        with open(os.path.join(self.path_to_saved, '62606aea-fc31-4706-8074-83a705fb29e5.json'), "r") as f:
            result = f.read()

        assert result is not None or ""

    def test_directory_creation_passes(self):
        folder = os.path.join(os.path.dirname(__file__), 'saved_instances/')

        if os.path.isdir(folder):
            os.rmdir(folder)

        self.wapp.start(address="127.0.0.1", port=8080)
        self.wapp.stop()

        assert os.path.isdir(folder)


class TestLoadObjectInstances:

    @pytest.fixture(autouse=True)
    def clean_files(self):
        folder = os.path.join(os.path.dirname(__file__), 'saved_instances/')

        if os.path.exists(folder):
            for f in os.listdir(folder):
                if os.path.isfile(os.path.join(folder, f)):
                    os.remove(os.path.join(folder, f))

    @pytest.fixture(autouse=True)
    def mock_server_sockets(self, mocker):
        from wappsto.connection.communication import ClientSocket
        mocker.patch.object(socket, 'socket', new=fake_socket.FakeSocket)
        mocker.patch.object(ClientSocket, 'ssl_wrap', new=fake_set)
        mocker.patch.object(ClientSocket, 'init_ok', new=lambda f: True)

    @classmethod
    def setup_class(self):
        one = 'test_JSON/62606aea-fc31-4706-8074-83a705fb29e5.json'
        two = 'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce.json'
        three = 'test_JSON/e9ec36e2-0bd5-4fef-8e7b-9c200e93ebce_invalid.json'
        self.json_value_one = os.path.join(os.path.dirname(__file__),one)
        self.json_ok_two_devices = os.path.join(os.path.dirname(__file__),two)
        self.json_incorrect = os.path.join(os.path.dirname(__file__), three)
        self.wapp = wappsto.Wappsto(json_file_name=self.json_value_one)
        self.path_to_saved = os.path.join(self.wapp.path_to_calling_file, 'saved_instances/')

    def test_load_files(self):

        self.wapp1 = wappsto.Wappsto(json_file_name=self.json_value_one)
        self.wapp1.start(address="127.0.0.1", port=8080)
        self.wapp1.stop()

        self.wapp2 = wappsto.Wappsto(load_from_state_file=True)
        self.wapp2.start(address="127.0.0.1", port=8080)

        assert self.wapp2.instance is not None

    def test_loading_empty_saved_instances_directory_errors(self, clean_files):
        with pytest.raises(ValueError):
            self.wapp_test = wappsto.Wappsto(load_from_state_file=True)

    @classmethod
    def teardown_method(self):
        for f in os.listdir(self.path_to_saved):
            if os.path.isfile(os.path.join(self.path_to_saved, f)):
                os.remove(os.path.join(self.path_to_saved, f))


class TestValueUnitTests:

    @classmethod
    def setup_method(self):
        self.parent_device = Mock(spec=Device)
        self.parent_device.uuid = "1"
        self.report_state = Mock(spec=State)
        self.report_state.uuid = "1.1.1"
        self.report_state.timestamp = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        self.control_state = Mock(spec=State)
        self.control_state.uuid = "1.1.2"
        self.rpc = Mock(spec=SeluxitRpc)
        self.value_number = wappsto.connection.network_classes.value.Value(
            self.parent_device, "1.1", "value_number", "number", "number",
            "rw", 0, 5, 1, 1, "cm", None, None, None, None,
        )

        self.value_blob = wappsto.connection.network_classes.value.Value(
            self.parent_device, "1.2", "value_blob", "blob", "blob",
            "rw", None, None, None, None, None, None, None, "utf-8", "15"
        )

        self.value_string = wappsto.connection.network_classes.value.Value(
            self.parent_device, "1.3", "value_string", "string", "string",
            "rw", None, None, None, None, None, "utf-8", "15", None, None
        )


    def test_value_set_period_report_state_not_exists(self):
        self.value_number.set_period(2)
        assert self.value_number.period is None

    def test_value_set_period_less_than_zero(self):
        self.value_number.set_period(-5)
        assert self.value_number.period is None

    def test_value_set_delta_without_report_state(self):
        self.value_number.set_delta(2)
        assert self.value_number.delta is None

    def test_value_set_delta_with_report_state(self):
        self.value_number.add_report_state(self.report_state)
        self.value_number.set_delta(2)
        assert self.value_number.delta == 2

    def test_value_set_delta_less_than_zero(self):
        self.value_number.add_report_state(self.report_state)
        self.value_number.set_delta(-1)
        assert self.value_number.delta is None

    def test_value_set_delta_for_not_number_type_value(self):
        self.value_blob.add_report_state(self.report_state)
        self.value_blob.set_delta(2)
        assert self.value_blob.delta is None

    def test_callback_is_not_set_after_initialization_blob(self):
        assert self.value_blob.callback is None

    def test_callback_is_not_set_after_initialization_number(self):
        assert self.value_number.callback is None

    def test_callback_is_not_set_after_initialization_string(self):
        assert self.value_string.callback is None

    def test_set_callback_not_callable(self):
        wappsto_errors_folder = wappsto.connection.network_classes.errors
        exc = wappsto_errors_folder.wappsto_errors.CallbackNotCallableException
        with pytest.raises(exc):
            self.value_number.set_callback("Function")

    def test_set_callback_callable(self):
        wappsto_errors_folder = wappsto.connection.network_classes.errors
        exc = wappsto_errors_folder.wappsto_errors.CallbackNotCallableException

        def callback_method(a,b):
            return a+b
        assert self.value_number.set_callback(callback=callback_method) is True

    def test_send_report_difference_bigger_than_delta(self):
        self.value_number.difference = 5
        self.value_number.delta = 2
        self.value_number.rpc = self.rpc
        self.value_number.delta_report = 1
        self.value_number.send_report_delta(self.report_state)
        no_send_logic_calls = self.value_number.rpc.get_rpc_state.call_count
        assert no_send_logic_calls is 1

    def test_send_report_difference_equals_delta(self):
        self.value_number.difference = 5
        self.value_number.delta = 5
        self.value_number.rpc = self.rpc
        self.value_number.delta_report = 1
        self.value_number.send_report_delta(self.report_state)
        no_send_logic_calls = self.value_number.rpc.get_rpc_state.call_count
        assert no_send_logic_calls is 1

    def test_send_report_difference_smaller_than_delta(self):
        self.value_number.difference = 3
        self.value_number.delta = 10
        self.value_number.rpc = self.rpc
        self.value_number.delta_report = 1
        self.value_number.send_report_delta(self.report_state)
        no_send_logic_calls = self.value_number.rpc.get_rpc_state.call_count
        assert no_send_logic_calls is 0

    def test_send_control_without_control_state_returns_false(self):
        result = self.value_number.send_control(2)
        assert result is False

    def test_send_control_exceed_allowed_above_number_value_number(self):
        self.value_number.add_control_state(self.control_state)
        result = self.value_number.send_control(10)
        assert result is False

    def test_send_control_exceed_allowed_below_number_value_number(self):
        self.value_number.add_control_state(self.control_state)
        result = self.value_number.send_control(-50)
        assert result is False

    def test_send_control_allowed_number_value_number(self):
        self.value_number.add_control_state(self.control_state)
        self.value_number.rpc = self.rpc
        self.value_number.send_control(2)
        no_send_logic_calls = self.value_number.rpc.get_rpc_state.call_count
        assert no_send_logic_calls is 1

    def test_send_control_string_max_is_none_value_string(self):
        self.value_string.add_control_state(self.control_state)
        self.value_string.rpc = self.rpc
        self.value_string.string_max = None
        self.value_string.send_control("Hello" * 30)
        no_send_logic_calls = self.value_string.rpc.get_rpc_state.call_count
        assert no_send_logic_calls is 1

    def test_send_control_blob_max_is_none_value_blob(self):
        self.value_blob.add_control_state(self.control_state)
        self.value_blob.rpc = self.rpc
        self.value_blob.blob_max = None
        self.value_blob.send_control("Hello" * 20)
        no_send_logic_calls = self.value_blob.rpc.get_rpc_state.call_count
        assert no_send_logic_calls is 1

    def test_send_control_string_max_is_15_value_valid_value_string(self):
        self.value_string.add_control_state(self.control_state)
        self.value_string.rpc = self.rpc
        self.value_string.send_control("Imagination")
        no_send_logic_calls = self.value_string.rpc.get_rpc_state.call_count
        assert no_send_logic_calls is 1

    def test_send_control_string_max_is_15_value_invalid_value_string(self):
        self.value_string.add_control_state(self.control_state)
        self.value_string.rpc = self.rpc
        self.value_string.send_control("Imagination" * 10)
        no_send_logic_calls = self.value_string.rpc.get_rpc_state.call_count
        assert no_send_logic_calls is 0

    def test_send_control_blob_max_is_15_value_valid_value_string(self):
        self.value_blob.add_control_state(self.control_state)
        self.value_blob.rpc = self.rpc
        self.value_blob.send_control("Imagination")
        no_send_logic_calls = self.value_blob.rpc.get_rpc_state.call_count
        assert no_send_logic_calls is 1


class TestRealServerConnection:

    @pytest.fixture(autouse=True)
    def clear_logs(self):
        open("server_logging", "w").close()

    @classmethod
    def setup_class(self):
        self.server = mock_wappsto_server.MockServer("127.0.0.1", 8080)
        self.server.threaded_run()
        one = 'test_JSON/62606aea-fc31-4706-8074-83a705fb29e5.json'
        self.json_value_one = os.path.join(os.path.dirname(__file__), one)
        self.wapp = wappsto.Wappsto(json_file_name=self.json_value_one)

    def test_connection_to_mock_server(self):
        self.wapp.start("127.0.0.1", 8080)
        result = False
        file = open("server_logging", "r")
        for line in file:
            if "Connected to server!" in line:
                result = True
        assert result is True
